
import selectors
import weakref

READ = selectors.EVENT_READ
WRITE = selectors.EVENT_WRITE

class Descriptor:
	def __init__(self, loop, fn, fd=None):
		self._loop_ref = weakref.ref(loop)
		self.fn = fn
		self.fd = fd

		self.has_run = False
		self.mask = None

	def get_loop(self):
		loop = self._loop_ref()
		assert loop
		return loop

class Handle:
	def __init__(self, loop, desc, do_unregister):
		self._desc_ref = weakref.ref(desc)
		self._do_unregister = do_unregister
		self._was_unregistered = False

	def unregister(self):
		if self._was_unregistered:
			return

		desc = self._desc_ref()
		assert desc
		self._do_unregister(desc.get_loop(), desc)
		self._was_unregistered = True

class EventLoop:
	@staticmethod
	def _unregister_observer(self, desc):
		self._observers.remove(desc)
		self._refcount -= 1

	@staticmethod
	def _unregister_file(self, desc):
		self._sel.unregister(desc.fd)
		self._refcount -= 1

	@staticmethod
	def _unregister_idle(self, desc):
		self._idle.remove(desc)
		self._refcount -= 1

	def __init__(self):
		self._active = True
		self._refcount = 0
		self._sel = selectors.DefaultSelector()

		self._observers = set()
		self._idle = set()

	def shutdown(self):
		print("Shutting down event loop")
		assert self._active
		self._active = False

	def register_observer(self, fn):
		desc = Descriptor(self, fn)
		self._observers.add(desc)
		self._refcount += 1
		return Handle(self, desc, EventLoop._unregister_observer)

	def register_file(self, fd, mask, fn):
		desc = Descriptor(self, fn, fd=fd)
		self._sel.register(fd, mask, desc)
		self._refcount += 1
		return Handle(self, desc, EventLoop._unregister_file)

	def register_idle(self, fn):
		desc = Descriptor(self, fn)
		self._idle.add(desc)
		self._refcount += 1
		return Handle(self, desc, EventLoop._unregister_idle)

	def run(self):
		assert self._active
		assert self._refcount

		while self._refcount:
			# List of pending descriptors.
			pd = []

			# Schedule all observers that did not run yet.
			if not self._active:
				for desc in self._observers:
					if desc.has_run:
						continue
					pd.append(desc)

			# Schedule all I/O events.
			timeout = None
			# Do not block if there are any pending descriptors.
			if pd:
				timeout = 0
			events = self._sel.select(timeout)
			for key, mask in events:
				desc = key.data
				desc.mask = mask
				pd.append(desc)

			# Schedule all idle callbacks.
			pd = list(self._idle) + pd

			# Finally, run all events.
			for desc in pd:
				desc.fn(desc)
				desc.has_run = True





import os
import selectors
import signal
import subprocess
import time

import yaml
import warnings

from .. import base
from .. import util

class Launcher:
	pass

def lock_run(run):
	util.try_mkdir(os.path.join(run.config.basedir, 'aux'))
	util.try_mkdir(os.path.join(run.config.basedir, 'output'))
	util.try_mkdir(run.experiment.aux_subdir)
	util.try_mkdir(run.experiment.output_subdir)

	# We will try to launch the experiment.
	# First, create a .lock file. If that is successful, we are the process that
	# gets to launch the experiment. Afterwards, concurrent access to our files
	# can be considered a bug (or deliberate misuse) and will lead to hard failures.
	try:
		lockfd = os.open(run.aux_file_path('lock'),
				os.O_RDONLY | os.O_CREAT | os.O_EXCL, mode=0)
	except FileExistsError:
		# TODO: Those warnings should be behind a flag.
#				print("Warning: .lock file exists for experiment '{}', instance '{}'".format(
#						exp.name, instance))
#				print("Either experiments are launched concurrently or the launcher crashed.")
		return False
	os.close(lockfd)
	return True

def create_run_file(run):

	# Create the .run file. This signals that the run has been submitted.
	with open(run.aux_file_path('run.tmp'), "w"):
		pass
	os.rename(run.aux_file_path('run.tmp'), run.aux_file_path('run'))

# Stores all information that is necessary to invoke a run.
# This is a view over a POD object which can be YAML-encoded and sent
# over a wire or stored into a file.
class RunManifest:
	def __init__(self, yml):
		self.yml = yml

	@property
	def base_dir(self):
		return self.yml['config']['base_dir']

	@property
	def instance_dir(self):
		return self.yml['config']['instance_dir']

	@property
	def revision(self):
		return self.yml['revision']

	@property
	def instance(self):
		return self.yml['instance']

	@property
	def instance_yml_name(self):
		return self.yml['instance_filename']

	@property
	def instance_extensions(self):
		return self.yml['instance_extensions']

	@property
	def instance_files(self):
		return self.yml['instance_files']

	@property
	def experiment(self):
		return self.yml['experiment']

	@property
	def repetition(self):
		return self.yml['repetition']

	@property
	def args(self):
		return self.yml['args']

	@property
	def environ(self):
		env_vars = self.yml['environ']
		for variant in self.yml['variants']:
			env_vars.update(variant['environ'].items())

		return env_vars

	@property
	def timeout(self):
		return self.yml['timeout']

	@property
	def aux_subdir(self):
		return base.get_aux_subdir(self.base_dir, self.experiment,
				[var_yml['name'] for var_yml in self.yml['variants']],
				self.revision)

	@property
	def output_subdir(self):
		return base.get_output_subdir(self.base_dir, self.experiment,
				[var_yml['name'] for var_yml in self.yml['variants']],
				self.revision)

	@property
	def workdir(self):
		return self.yml['workdir']

	@property
	def builds(self):
		return self.yml['builds']

	@property
	def output_extensions(self):
		return self.yml['output_extensions']

	@property
	def stdout(self):
		return self.yml['stdout']

	@property
	def variants(self):
		return [v['name'] for v in self.yml['variants']]

	def aux_file_path(self, ext):
		return os.path.join(self.aux_subdir,
				base.get_aux_file_name(ext, self.instance, self.repetition))

	def output_file_path(self, ext):
		return os.path.join(self.output_subdir,
				base.get_output_file_name(ext, self.instance, self.repetition))

	def get_extra_args(self):
		extra_args = []
		for var_yml in self.yml['variants']:
			extra_args.extend(var_yml['extra_args'])
		return extra_args

	def get_paths(self):
		paths = []
		for build_yml in self.yml['builds'].values():
			paths.append(os.path.join(build_yml['prefix'], 'bin'))
		return paths

	def get_ldso_paths(self):
		paths = []
		for build_yml in self.yml['builds'].values():
			paths.append(os.path.join(build_yml['prefix'], 'lib64'))
			paths.append(os.path.join(build_yml['prefix'], 'lib'))
		return paths

	def get_python_paths(self):
		paths = []
		for build_yml in self.yml['builds'].values():
			for export in build_yml['exports_python']:
				paths.append(os.path.join(build_yml['prefix'], export))
		return paths

	def get_source_dir_for(self, build_name):
		if build_name not in self.yml['builds']:
			raise RuntimeError("The experiment '{}' does not use the build '{}'".format(self.yml['experiment'], build_name))
		return self.yml['builds'][build_name]['source']

	def get_compile_dir_for(self, build_name):
		if build_name not in self.yml['builds']:
			raise RuntimeError("The experiment '{}' does not use the build '{}'".format(self.yml['experiment'], build_name))
		return self.yml['builds'][build_name]['compile']

	def get_prefix_dir_for(self, build_name):
		if build_name not in self.yml['builds']:
			raise RuntimeError("The experiment '{}' does not use the build '{}'".format(self.yml['experiment'], build_name))
		return self.yml['builds'][build_name]['prefix']

def compile_manifest(run):
	exp = run.experiment

	# Perform a DFS to discover all used builds.
	recursive_builds = []
	builds_visited = set()

	for name in exp.info.used_builds:
		assert name not in builds_visited
		build = run.config.get_build(name, exp.revision)
		recursive_builds.append(build)
		builds_visited.add(name)

	i = 0  # Need index-based loop as recursive_builds is mutated in the loop.
	while i < len(recursive_builds):
		build = recursive_builds[i]
		for req_name in build.info.requirements:
			if req_name in builds_visited:
				continue
			req_build = run.config.get_build(req_name, exp.revision)
			recursive_builds.append(req_build)
			builds_visited.add(req_name)
		i += 1

	instance_files = None
	instance_extensions = None
	if run.instance.has_multi_files:
		instance_files = run.instance.filenames
	elif run.instance.has_multi_ext:
		instance_extensions = run.instance.extensions

	builds_dict = {}
	for build in recursive_builds:
		builds_dict[build.name] = {
			'source': build.source_dir if build.revision.is_dev_build else build.clone_dir,
			'compile': build.compile_dir,
			'prefix': build.prefix_dir,
			'exports_python': build.info.exports_python,
			'extra_paths': util.ensure_list_type(build.info.extra_paths)
		}

	# Collect extra arguments from variants
	variants_yml = []
	for variant in exp.variation:
		environ = {}
		if 'environ' in variant.variant_yml:
			for (k, v) in variant.variant_yml['environ'].items():
				environ[k] = str(v)
		variants_yml.append({
			'name': variant.name,
			'extra_args': variant.variant_yml.get('extra_args', []),
			'environ': environ
		})

	environ = {}
	for (k, v) in exp.info.environ.items():
		environ[k] = str(v)

	stdout = None
	if exp.info.stdout is not None:
		stdout = exp.info.stdout
	elif exp.info.output == 'stdout':
		stdout = 'out'

	return RunManifest({
		'config': {
			'base_dir': run.config.basedir,
			'instance_dir': run.config.instance_dir()
		},
		'experiment': exp.name,
		'variants': variants_yml,
		'revision': exp.revision.name if exp.revision else None,
		'instance': run.instance.shortname,
		'instance_filename': run.instance.yml_name,
		'instance_extensions': instance_extensions,
		'instance_files': instance_files,
		'repetition': run.repetition,
		'builds': builds_dict,
		'args': exp.info.args,
		'timeout': float(exp.info.timeout) if exp.info.timeout is not None else exp.info.timeout,
		'environ': environ,
		'output_extensions': exp.info.output_extensions,
		'stdout': stdout,
		'workdir': exp.info.workdir
	})

def invoke_run(manifest):

	def get_qualified_output_file(ext):
		if ext not in manifest.output_extensions:
			raise RuntimeError(
				f"Unexpected output extension for experiment '{manifest.experiment}': .{ext}\n"
			)
		return manifest.output_file_path(ext)

	(stdout_pipe, stdout) = (None, None)
	if manifest.stdout is not None:
		stdout_path = get_qualified_output_file(manifest.stdout)

		# We do not actually need to write anything to the output file.
		# However, we might want to pipe experimental output to it.
		with open(stdout_path, 'w') as f:
			stdout = os.dup(f.fileno())
	else:
		stdout_path = manifest.aux_file_path('stdout')
		(stdout_pipe, stdout) = os.pipe()
		os.set_blocking(stdout_pipe, False)

	# Create all the other the output files.
	# The creation of the output file with extension '.out' signals that the run has been started.
	diff_set = {manifest.stdout} if manifest.stdout is not None else set()
	for ext in manifest.output_extensions - diff_set:
		util.touch(manifest.output_file_path(ext))

	# Create the error file.
	(stderr_pipe, stderr) = os.pipe()
	os.set_blocking(stderr_pipe, False)

	def get_qualified_filename(identifier):
		if identifier.isdigit():
			identifier = int(identifier)
			if manifest.instance_files is None:
				raise RuntimeError(
					f"Instance '{manifest.instance}' does not have any files specified in the experiments.yml"
				) from None
			if len(manifest.instance_files) <= identifier:
				raise IndexError('File index out of range: {}'.format(identifier))
			return ''.join([manifest.instance_dir, '/', manifest.instance_files[identifier]])
		else:
			if manifest.instance_extensions is None:
				raise RuntimeError(
					f"Instance '{manifest.instance}' does not have any extensions specified in the experiments.yml"
				) from None
			if identifier not in manifest.instance_extensions:
				raise RuntimeError(
					f"Unexpected file extension for instance '{manifest.instance}': .{identifier}"
				) from None
			return ''.join([manifest.instance_dir, '/', manifest.instance_yml_name, '.', identifier])

	def substitute(p):
		if p == 'INSTANCE':
			return manifest.instance_dir + '/' + manifest.instance_yml_name
		elif p.startswith('INSTANCE:'):
			return get_qualified_filename(p.split(':')[1])
		elif p == 'REPETITION':
			return str(manifest.repetition)
		elif p == 'OUTPUT':
			return stdout_path
		elif p.startswith('OUTPUT:'):
			return get_qualified_output_file(p.split(':')[1])
		elif p.startswith('SOURCE_DIR_FOR:'):
			return manifest.get_source_dir_for(p.split(':')[1])
		elif p.startswith('COMPILE_DIR_FOR:'):
			return manifest.get_compile_dir_for(p.split(':')[1])
		elif p.startswith('PREFIX_DIR_FOR:'):
			return manifest.get_prefix_dir_for(p.split(':')[1])
		elif p == 'OUTPUT_SUBDIR':
			return manifest.output_subdir
		else:
			return None

	def substitute_list(p):
		if p == 'EXTRA_ARGS':
			return manifest.get_extra_args()
		else:
			return None

	cmd = util.expand_at_params(manifest.args, substitute, listfn=substitute_list)

	# Build the environment.
	def prepend_env(var, items):
		if var in os.environ:
			return ':'.join(items) + ':' + os.environ[var]
		return ':'.join(items)

	def substitute_extra_paths(p):
		if p in ['THIS_CLONE_DIR', 'THIS_SOURCE_DIR']:
			return build_yml['source']
		elif p == 'THIS_COMPILE_DIR':
			return build_yml['compile']
		elif p == 'THIS_PREFIX_DIR':
			return build_yml['prefix']
		elif p.startswith('SOURCE_DIR_FOR:'):
			return manifest.get_source_dir_for(p.split(':')[1])
		elif p.startswith('COMPILE_DIR_FOR:'):
			return manifest.get_compile_dir_for(p.split(':')[1])
		elif p.startswith('PREFIX_DIR_FOR:'):
			return manifest.get_prefix_dir_for(p.split(':')[1])
		else:
			return None
	environ = os.environ.copy()

	extra_paths = manifest.get_paths()
	for build_yml in manifest.builds.values():
		extra_paths.extend(util.expand_at_params(build_yml['extra_paths'], substitute_extra_paths))

	environ['PATH'] = prepend_env('PATH', extra_paths)
	environ['LD_LIBRARY_PATH'] = prepend_env('LD_LIBRARY_PATH', manifest.get_ldso_paths())
	environ['PYTHONPATH'] = prepend_env('PYTHONPATH', manifest.get_python_paths())
	environ.update(manifest.environ)

	# Dumps data from an FD to the FS.
	# Creates the output file only if something is written.
	class LazyWriter:
		def __init__(self, fd, path):
			self._fd = fd
			self._path = path
			self._out = None

		def progress(self):
			# Specify some chunk size to avoid reading the whole pipe at once.
			chunk = os.read(self._fd, 16 * 1024)
			if not len(chunk):
				return False

			if self._out is None:
				self._out = open(self._path, "wb")
			self._out.write(chunk)
			self._out.flush()
			return True

		def close(self):
			if self._out is not None:
				self._out.close()

	start = time.perf_counter()
	cwd = (util.expand_at_params(manifest.workdir, substitute)
			if manifest.workdir is not None else manifest.base_dir)
	child = subprocess.Popen(cmd, cwd=cwd, env=environ,
			stdout=stdout, stderr=stderr)
	os.close(stderr)
	sel = selectors.DefaultSelector()

	if manifest.stdout is None:
		os.close(stdout)
		stdout_writer = LazyWriter(stdout_pipe, manifest.aux_file_path('stdout'))
		sel.register(stdout_pipe, selectors.EVENT_READ, stdout_writer)
	stderr_writer = LazyWriter(stderr_pipe, manifest.aux_file_path('stderr'))
	sel.register(stderr_pipe, selectors.EVENT_READ, stderr_writer)

	# Wait until the run program finishes.
	while True:
		if child.poll() is not None:
			break

		elapsed = time.perf_counter() - start
		if manifest.timeout is not None and elapsed > manifest.timeout:
			# Programs can catch the SIGXCPU signal and keep running.
			# Thus, the kill command ensures that the time limit is respected (with a grace period).
			if elapsed > manifest.timeout + base.TIMEOUT_GRACE_PERIOD:
				child.kill()
			else:
				child.send_signal(signal.SIGXCPU)

		# Consume any output that might be ready.
		events = sel.select(timeout=1)
		for (sk, mask) in events:
			if not sk.data.progress():
				sel.unregister(sk.fd)

	# Consume all remaining output.
	while True:
		events = sel.select(timeout=0)
		for (sk, mask) in events:
			if not sk.data.progress():
				sel.unregister(sk.fd)
		if not events:
			break
	if manifest.stdout is None:
		stdout_writer.close()
		os.close(stdout_pipe)
	stderr_writer.close()
	os.close(stderr_pipe)
	runtime = time.perf_counter() - start

	# Collect the status information.
	status = None
	sigcode = None
	if child.returncode < 0:  # Killed by a signal?
		sigcode = signal.Signals(-child.returncode).name
	else:
		status = child.returncode
	did_timeout = manifest.timeout is not None and runtime > manifest.timeout

	# Create the status file to signal that we are finished.
	status_dict = {'timeout': did_timeout, 'walltime': runtime,
			'status': status, 'signal': sigcode}
	with open(manifest.output_file_path('status.tmp'), "w") as f:
		yaml.dump(status_dict, f)
	os.rename(manifest.output_file_path('status.tmp'), manifest.output_file_path('status'))




import errno
import os
import re
import shutil
import sys
import yaml
import json
from jsonschema import Draft7Validator
import warnings

def expand_at_params(s, fn, listfn=None):
	def subfn(m):
		result = fn(m.group(1))
		if result is None:
			raise RuntimeError("Unexpected @-parameter '{}' in {}".format(m.group(1), s))
		return result

	if isinstance(s, list):
		seq = [ ]
		for item in s:
			# Try to replace the whole item.
			if listfn is not None:
				m = re.fullmatch(r'@(\w+)@', item)
				if m:
					result = listfn(m.group(1))
					if result is not None:
						seq.extend(result)
						continue

			seq.append(re.sub(r'@(\w+(:[^@]+)?)@', subfn, item))
		return seq
	else:
		assert isinstance(s, str)
		return re.sub(r'@(\w+(:[^@]+)?)@', subfn, s)

def try_mkdir(path):
	try:
		os.mkdir(path)
	except OSError as error:
		if error.errno != errno.EEXIST:
			raise

def try_rmfile(path):
	try:
		os.remove(path)
	except OSError as error:
		if error.errno != errno.ENOENT:
			raise

def try_rmtree(path):
	try:
		shutil.rmtree(path)
	except OSError as error:
		if error.errno != errno.ENOENT:
			raise

def touch(path):
	with open(path, 'w'):
		pass

def yaml_to_string(yml):
	return yaml.dump(yml, Dumper=yaml.SafeDumper)

def write_yaml_file(f, yml):
	return yaml.dump(yml, f, Dumper=yaml.SafeDumper)

def yaml_from_string(string):
	from simexpal.base import YmlLoader
	return yaml.load(string, Loader=YmlLoader)

def read_yaml_file(f):
	from simexpal.base import YmlLoader
	return yaml.load(f, Loader=YmlLoader)

def read_setup_file(setup_file):
	from simexpal.base import YmlLoader

	with open(setup_file, 'r') as f:
		setup_dict = yaml.load(f, Loader=YmlLoader)
	return setup_dict

def validate_setup_file(setup_file):
	""" Reads, validates and sanitizes the setup file
	"""

	def _validate_dict(dictionary, source, schema_path):
		with open(schema_path, 'r') as f:
			schema = json.load(f)

		validator = Draft7Validator(schema)
		validation_errors = list(validator.iter_errors(dictionary))

		if len(validation_errors) > 0:
			for err in validation_errors:
				err_source = "[{}]".format("][".join(str(x) for x in err.absolute_path))

				print("simexpal: Validation error in {} at {}:\n{}\n{}".format(
					source, err_source, err.instance, err.message), file=sys.stderr, end="\n\n")
			sys.exit(1)

	cur_file_path = os.path.abspath(os.path.dirname(__file__))

	exp_yml_dict = read_setup_file(setup_file)
	schema_path = os.path.join(cur_file_path, "schemes", "experiments.json")
	_validate_dict(exp_yml_dict, "experiments.yml", schema_path)

	try:
		launchers_yml_dict = read_setup_file(os.path.expanduser('~/.simexpal/launchers.yml'))
		schema_path = os.path.join(cur_file_path, "schemes", "launchers.json")
		_validate_dict(launchers_yml_dict, "launchers.yml", schema_path)
	except FileNotFoundError:
		pass

	for exp in exp_yml_dict.get('experiments', []):
		if exp.get('output', None) == 'stdout':
			msg = "Specifying the stdout path via 'output: stdout' is deprecated and will be removed in future " \
					"versions. Use 'stdout: out' instead."
			warnings.warn(msg, DeprecationWarning)

			break

	if 'instdir' not in exp_yml_dict:
		exp_yml_dict['instdir'] = './instances'

	return exp_yml_dict

def compute_network_size(path, out):
	import networkit as nk
	try:
		g = nk.readGraph(path, nk.Format.EdgeList,
				separator=' ', firstNode=0, continuous=False, directed=False)
	except Exception: # Exception due the attempt of reading a non-network file
		return
	data = {
		'n': g.numberOfNodes(),
		'm': g.numberOfEdges()
	}
	yaml.dump(data, out, default_flow_style=False)

def ensure_list_type(arg):
	if isinstance(arg, list):
		return arg
	assert isinstance(arg, str)
	return [arg]

def read_file(path):
	try:
		f = open(path, 'r')
	except FileNotFoundError:
		return ''
	else:
		with f:
			return f.read()




from enum import Enum
import functools
import os
import socket

from . import launch
from .launch import common
from . import util
from . import evloop

class _State(Enum):
	REQUEST = 0
	DONE = 1

class _Queue:
	@staticmethod
	def handle_sock(self, descriptor):
		accepted, _ = self.sock.accept()
		con = _Connection(self, accepted)
		con.run(descriptor.get_loop())

	@staticmethod
	def observe_loop(self, descriptor):
		self._handle.unregister()
		self._observer.unregister()

		if self.path is not None:
			os.unlink(self.path)

	def __init__(self, sock, path=None):
		self.sock = sock
		self.path = path
		self._handle = None
		self._observer = None

	def run(self, loop):
		self._handle = loop.register_file(self.sock, evloop.READ,
				functools.partial(_Queue.handle_sock, self))
		self._observer = loop.register_observer(functools.partial(_Queue.observe_loop, self))
		self.sock.listen()

	# Dispatch a request and return a response.
	# This function is the main workhorse of the queue server.
	def dispatch(self, loop, req):
		if req['action'] == 'stop':
			loop.shutdown()
		else:
			assert req['action'] == 'launch'
			manifest = launch.common.RunManifest(req['manifest'])

			print("Processing experiment '{}', instance '{}'".format(
					manifest.experiment, manifest.instance))
			launch.common.invoke_run(manifest)

class _Connection:
	@staticmethod
	def handle_sock(self, descriptor):
		assert self.state == _State.REQUEST

		data = self.sock.recv(4096)
		if data:
			self.recv_buffer += data
			return

		req = util.yaml_from_string(self.recv_buffer.decode())
		resp = self.queue.dispatch(descriptor.get_loop(), req)
		self.state = _State.DONE
		self._handle.unregister()

	@staticmethod
	def observe_loop(self, descriptor):
		self._handle.unregister()
		self._observer.unregister()

	def __init__(self, queue, sock):
		self.queue = queue
		self.sock = sock
		self.recv_buffer = bytes()
		self.state = _State.REQUEST
		self._handle = None
		self._observer = None

	def run(self, loop):
		self._handle = loop.register_file(self.sock, evloop.READ,
				functools.partial(_Connection.handle_sock, self))
		self._observer = loop.register_observer(functools.partial(_Connection.observe_loop, self))

def run_queue(loop, sockfd=None, force=False):
	if sockfd is not None:
		serve_sock = socket.socket(fileno=sockfd)
		queue = _Queue(serve_sock)
	else:
		sockpath = os.path.expanduser('~/.extlq.sock')
		serve_sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		if force:
			util.try_rmfile(sockpath)
		serve_sock.bind(sockpath)
		queue = _Queue(serve_sock, path=sockpath)

	queue.run(loop)
	print('Serving on {}'.format(serve_sock.getsockname()))

def sendrecv(m):
	sockpath = os.path.expanduser('~/.extlq.sock')
	s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
	s.connect(sockpath)
	s.send(util.yaml_to_string(m).encode())
	s.close()

def stop_queue():
	sendrecv({
		'action': 'stop'
	})




from . import common
from .. import queuesock

class QueueLauncher(common.Launcher):
	def submit(self, config, run):
		if not common.lock_run(run):
			return
		common.create_run_file(run)

		print("Launching experiment '{}', instance '{}' on local machine".format(
				run.experiment.name, run.instance.shortname))

		queuesock.sendrecv({
			'action': 'launch',
			'manifest': common.compile_manifest(run).yml
		})





import collections
import os
import subprocess
import sys
import tempfile

from .. import util
from . import common

script_template = '''#!/bin/sh
@SIMEX@ internal-invoke @MODE@ @SPECFILE@
'''

class SlurmLauncher(common.Launcher):
	def __init__(self, queue):
		self.queue = queue

	def submit(self, cfg, run):
		self._do_submit(cfg, run.experiment, [run])

	def submit_multiple(self, cfg, runs):
		groups = collections.defaultdict(list)
		for run in runs:
			groups[run.experiment].append(run)
		for grp_exp, grp_runs in groups.items():
			self._do_submit(cfg, grp_exp, grp_runs)

	def _do_submit(self, cfg, experiment, runs):
		util.try_mkdir(os.path.join(cfg.basedir, 'aux'))
		util.try_mkdir(os.path.join(cfg.basedir, 'aux/_slurm'))

		# Lock the runs to make sure that we do not submit runs twice.
		locked = []
		for run in runs:
			if not common.lock_run(run):
				continue
			locked.append(run)

		if not locked:
			return
		use_array = len(locked) > 1

		# Build the specfile.
		if not use_array:
			specs = {
				'manifest': common.compile_manifest(locked[0]).yml
			}
		else:
			specs = {
				'manifests': [common.compile_manifest(run).yml for run in locked]
			}

		(specfd, specfile) = tempfile.mkstemp(prefix='', suffix='-spec.yml',
				dir=os.path.join(cfg.basedir, 'aux/_slurm'))
		with os.fdopen(specfd, 'w') as f:
			util.write_yaml_file(f, specs)

		# Expand the script that is passed to sbatch.
		def substitute(p):
			if p == 'SIMEX':
				return os.path.abspath(sys.argv[0])
			elif p == 'SPECFILE':
				return specfile
			elif p == 'MODE':
				return '--slurm-array' if use_array else '--slurm'
			else:
				return None

		sbatch_script = util.expand_at_params(script_template, substitute)

		ps = experiment.effective_process_settings
		ts = experiment.effective_thread_settings

		# Build the sbatch command to run the script.
		# TODO: Support multiple queues
		sbatch_args = ['sbatch', '-J', experiment.display_name]
		if self.queue:
			sbatch_args += ['-p', self.queue]
		if ps and ps['num_nodes']:
			sbatch_args += ['-N', str(ps['num_nodes'])]
		if ps and ps['procs_per_node']:
			sbatch_args += ['--ntasks-per-node', str(ps['procs_per_node'])]
		if ts and ts['num_threads']:
			sbatch_args += ['-c', str(ts['num_threads'])]
		log_pattern = '%A-%a' if use_array else '%A'
		sbatch_args.extend(['-o', os.path.join(cfg.basedir, 'aux/_slurm/' + log_pattern + '.out'),
				'-e', os.path.join(cfg.basedir, 'aux/_slurm/' + log_pattern + '.err')])

		if use_array:
			sbatch_args.append('--array=0-' + str(len(locked) - 1))

		# Add custom sbatch parameters of the user.
		slurm_args = util.ensure_list_type(experiment.effective_slurm_args)
		sbatch_args.extend(slurm_args)

		# Finally start the run.
		for run in locked:
			if self.queue:
				print("Submitting experiment '{}', instance '{}' to slurm partition '{}'".format(
						run.experiment.name, run.instance.shortname, self.queue))
			else:
				print("Submitting experiment '{}', instance '{}' to default slurm partition".format(
						run.experiment.name, run.instance.shortname))

		process = subprocess.Popen(sbatch_args, stdin=subprocess.PIPE)
		process.communicate(sbatch_script.encode())  # Assume UTF-8 encoding here.
		assert process.returncode == 0

		for run in locked:
			common.create_run_file(run)

